/* $Id: connection.cpp 453 2011-02-27 10:06:27Z oh2gve $
 *
 * Copyright 2007 Antti Seppänen
 * Copyright 2007, 2008, 2009, 2010, 2011 Tapio Aaltonen
 *
 * 
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/

// own stuff.
#include "connection.h"
#include "packetlog.h"
#include "connection.h"
#include "cfg.h"
#include "packet.h"

// std stuff.
#include <queue>
using std::queue;
#include <string>
using std::string;

// wx stuff.
#include <wx/thread.h>



Connection::Connection(Cfg::ConnectionState const& state, char const& conId, Packetlog* const& packetlog, DupeFilter* const& dupefilter):
wxThread(wxTHREAD_JOINABLE),
state_(state), conId_(conId), packetlog_(packetlog), dupefilter_(dupefilter), mutex_(),
isStarted_(false), hasFatalError_(false), keepgoing_(true), inputQueue_(), outputQueue_()
{
}


Connection::~Connection()
{
	// Flush queues.
	while ( !inputQueue_.empty() )
	{
		delete inputQueue_.front();
		inputQueue_.pop();
	}   

	while ( !outputQueue_.empty() )
	{
		delete outputQueue_.front();
		outputQueue_.pop();
	}   
}


// PUBLIC


void Connection::waitForInit()
{
	// Wait for the connection thread to actually start and lock it's mutex.
	while ( !isStarted_ )
	{
		Yield();
	}
	
	// Wait here until thread completes startup routines.
	mutex_.Lock();
	
	// Okay, it's up.
	mutex_.Unlock();
}


bool Connection::operator!()
{
	mutex_.Lock();
	bool retval = hasFatalError_;
	mutex_.Unlock();
	
	return retval;
}


void Connection::terminate()
{
	mutex_.Lock();
	keepgoing_ = false;
	mutex_.Unlock();
}


void Connection::write(Packet* const& data)
{
	// Just delete the packet if we can't write.
	if ( state_ == Cfg::READONLY )
	{
		delete data;
		return;
	}
	
	// Set our ID as output.
	data->setOutputId(conId_);
	
	// Put the data to outputQueue to wait for sending.
	mutex_.Lock();
	if ( !hasFatalError_) 
	{
		outputQueue_.push(data);
	}
	else
	{
		delete data;
	}
	mutex_.Unlock();
}


Packet* Connection::read()
{
	Packet* retval = NULL;
	
	mutex_.Lock();
	if ( !hasFatalError_ )
	{
		// Check for data in queue.
		if ( !inputQueue_.empty() )
		{
			// Return next packet.
			retval = inputQueue_.front();
			inputQueue_.pop();
		}
	}
	mutex_.Unlock();

	return retval;
}


// PROTECTED


void Connection::threadStarted()
{
	// This thread is now running, so it can lock it's mutex to hold main
	// thread at waitForInit() until connection has been opened or opening
	// has failed.
	mutex_.Lock();
	
	// Tell main thread that it can begin to wait for our mutex to unlock.
	isStarted_ = true;
}


void Connection::startupDone()
{
	// Unlock the mutex so that main thread can check our status.
	mutex_.Unlock();
}


void Connection::weHasError()
{
	mutex_.Lock();
	hasFatalError_ = true;
	mutex_.Unlock();
}


bool Connection::keepGoing()
{
	mutex_.Lock();
	bool retval = keepgoing_ && !hasFatalError_;
	mutex_.Unlock();
	
	return retval;
}


void Connection::inputPush(AprsPacket* const& data)
{
	// If we are not supposed to input, just delete the packet.
	if ( state_ == Cfg::WRITEONLY )
	{
		delete data;
		return;
	}
	
	// Check for errors.
	string errmsg;
	if ( data->error(errmsg) )
	{
		Packet* packet = new Packet(conId_, '_', data, 'e');
		packetlog_->write(packet);
		return;
	}
	
	// Try to get third-party packet from the packet.
	AprsPacket* actual = data;
	while ( AprsPacket* nextresult = actual->getCarriedThirdParty() )
	{
		// Remove temporary packet.
		if ( actual != data )
		{
			delete actual;
		}
		
		// Update actual.
		actual = nextresult;
	}

	// Convert to packet.
	Packet* packet = new Packet(conId_, '_', data, '-');
	
	// Set actual packet, if it was different from the one seen on HW.
	if ( actual != data )
	{
		packet->setAprsPacket(actual);
	}
	
	// Check for dupeness.
	if ( dupefilter_->isDupe(actual) )
	{
		packet->setFlag('d');
		packetlog_->write(packet);
		return;
	}
	
	// Check if can be gated.
	if ( !actual->canBeGated() )
	{
		packetlog_->write(packet);
		return;
	}
	
	// Push packet to input buffer.
	mutex_.Lock();
	inputQueue_.push(packet);
	mutex_.Unlock();
}


Packet* Connection::outputPop()
{
	Packet* retval = NULL;
	
	// Check for packets from main.
	mutex_.Lock();
	if ( !outputQueue_.empty() )
	{
		// Get the pointer from output buffer.
		retval = outputQueue_.front();

		// Remove the element from the buffer.
		outputQueue_.pop();
	}
	mutex_.Unlock();
	
	return retval;
}
