/* $Id: main.cpp 448 2011-02-26 07:26:08Z oh2gve $
 * 
 * Copyright 2007, 2008, 2009, 2010 Tapio Aaltonen
 * 
 * 
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/


// Build config.
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// own stuff.
#include "main.h"
#include "../cfg.h"

// std stuff.
#include <csignal>
#include <string>
using std::string;
#include <vector>
using std::vector;
#include <iomanip>
using std::setw;
#include <map>
using std::map;
using std::make_pair;

// non-std stuff.
#include <fap.h>

// wx stuff.
#include <wx/log.h>
#include <wx/utils.h>
#include <wx/cmdline.h>
#include <wx/string.h>



AprsApp::AprsApp():
iAmDaddy_(false), gotSignal_(0), nullFd_(0), daemonMode_(false), pidFile_(NULL),
configFileName_(), logFile_(NULL), packetlog_(NULL), dupefilter_(NULL), beacon_(NULL),
configuration_(), uplink_(NULL), connections_()
{
}


AprsApp::~AprsApp()
{
}


// Our signal handler to be bound into OS.
void handleSignal(int num)
{
	AprsApp* tmp = dynamic_cast<AprsApp*>(wxTheApp);
	if ( tmp != NULL )
	{
		tmp->gotSignal(num);
	}
}


// COMMAND LINE PARSING

/// Command line options.
/**
 * Recognized command line options are listed here.
 */
wxCmdLineEntryDesc const CMDLINE[] =
{
	{ wxCMD_LINE_SWITCH, wxT("h"), wxT("help"), wxT("Show this help text."), wxCMD_LINE_VAL_STRING, wxCMD_LINE_OPTION_HELP },
	{ wxCMD_LINE_SWITCH, wxT("d"), wxT("daemon"), wxT("Run in background, not in terminal."), wxCMD_LINE_VAL_STRING, wxCMD_LINE_PARAM_OPTIONAL },
	{ wxCMD_LINE_SWITCH, wxT("v"), wxT("version"), wxString(wxT("Prints version of ")) + wxString(wxT(PACKAGE)) + wxString(wxT(" and quits.")), wxCMD_LINE_VAL_STRING, wxCMD_LINE_PARAM_OPTIONAL },
	{ wxCMD_LINE_PARAM,  wxT("c"), wxT("conf"), wxT("configfile"), wxCMD_LINE_VAL_STRING, wxCMD_LINE_PARAM_OPTIONAL },
	// Last line just terminates this list.
	{ wxCMD_LINE_NONE }
};

void AprsApp::OnInitCmdLine(wxCmdLineParser& parser)
{
	parser.SetDesc(CMDLINE);
}

bool AprsApp::OnCmdLineParsed(wxCmdLineParser& parser)
{
	// Check if it's version they want.
	if ( parser.Found(wxT("version")) )
	{
		wxLogMessage(wxString(wxT(PACKAGE)) + wxString(wxT(" v")) + wxString(wxT(PACKAGE_VERSION)));
		return false;
	}
	
	// Check if we're asked to go background.
	if ( parser.Found(wxT("daemon")) )
	{
		daemonMode_ = true;
	}
	
	// Look for config file name.
	if( parser.GetParamCount() > 0 )
	{
		configFileName_ = parser.GetParam();
	}
	else
	{
		// Not specified, let's set a default.
		configFileName_ = wxT(SYSCONFDIR);
		configFileName_ += wxT("/");
		configFileName_ += wxT(PACKAGE);
		configFileName_ += wxT(".conf");
	}
	
	return true;
}

bool AprsApp::OnInit()
{
	// Parse cmd line. Possibly does something else.
	if ( !wxAppConsole::OnInit() )
	{
		return false;
	}
	
	wxLogDebug(wxT("Main thread started."));
		
	// Fork into daemon process, if requested.
	if ( daemonMode_ )
	{
		int pid = fork();
		if ( pid == 0 )
		{
			// Child process goes here.
		}
		else if ( pid < 0 )
		{
			// Daddy process goes here on failure (no child).
			wxLogError(wxT("Can't start daemon process."));
			return false;
		}
		else
		{
			// Daddy process goes here on success.
			iAmDaddy_ = true;
			return true;
		}
	}

	// Initialize APRS parser.
	fap_init();
	
	// Read configuration from file.
	if ( !readCfg(configFileName_, configuration_) )
	{
		wxLogError(wxT("There was an error while reading ") + configFileName_ + wxT(", can't start."));
		fap_cleanup();
		return false;
	}
	
	// Initialize null file descriptor for stdio redirection.
	nullFd_ = open("/dev/null", O_RDONLY);
	
	// Create pid file.
	if ( !createPidFile() )
	{
		fap_cleanup();
		return false;
	}

	// If in daemon mode, move logging from stdio to a file.
	if ( daemonMode_ && !openLog() )
	{
		removePidFile();
		fap_cleanup();
		return false;
	}
	
	// Start assistant threads.
	if ( !startAssistants() )
	{
		closeLog();
		removePidFile();
		fap_cleanup();
		return false;
	}
		
	// Open the devices.
	if ( !openConnections() )
	{
		closeConnections();
		stopAssistants();
		closeLog();
		removePidFile();
		fap_cleanup();
		return false;
	}
	
	// Connect our signal handler to some signals.
	signal( SIGTERM, handleSignal );
	signal( SIGINT, handleSignal );
	signal( SIGHUP, handleSignal );
	
	// All systems are go.
	return true;
}



void AprsApp::gotSignal(int num)
{
	gotSignal_ = num;
}



bool AprsApp::reopenLog()
{
	if ( logFile_ != NULL )
	{
	closeLog();
	return openLog(false);
	}
	
	return true;
}


bool AprsApp::createPidFile()
{
	pidFile_ = new wxFile;
	wxString name = configuration_.rundir;
	name += wxT("/");
	name += wxT(PACKAGE);
	name += wxT(".pid");

	// Check if pidfile already exists.
	if ( pidFile_->Exists(name.GetData()) )
	{
		wxString msg = wxT("Error: another instance of ");
		msg += wxT(PACKAGE);
		msg += wxT(" seems to be running.");
		wxLogMessage(msg);

		delete pidFile_;
		pidFile_ = NULL;
		return false;
	}
	
	// Create pid file.
	if ( !pidFile_->Create(name.GetData()) )
	{
		wxLogMessage(wxT("Error: failed to create pid file."));
		delete pidFile_;
		pidFile_ = NULL;
		return false;
	}
	
	// Get our pid.
	wxString pid = wxT("");
	pid << wxGetProcessId();
	pid += wxT("\n");

	// Write our pid to pidfile.
	if ( !pidFile_->Write(pid) )
	{
		wxLogMessage(wxT("Error: cannot write to pid file ") + name + wxT("."));
		removePidFile();
		return false;
	}
	
	return true;
}

void AprsApp::removePidFile()
{
	if ( pidFile_ )
	{
		pidFile_->Close();
		delete pidFile_;
		pidFile_ = NULL;

		wxString name = configuration_.rundir;
		name += wxT("/");
		name += wxT(PACKAGE);
		name += wxT(".pid");

		// We just blindly try to remove the file.
		remove(name.ToAscii());
	}
}


int AprsApp::OnExit()
{
	closeConnections();
	stopAssistants();
	closeLog();
	removePidFile();
	fap_cleanup();
	close(nullFd_);
	wxLogDebug(wxT("Main thread is terminating."));
	return wxAppConsole::OnExit();
}


// Let the magic happen.
IMPLEMENT_APP_CONSOLE(AprsApp)
