/* $Id: serialport.cpp 461 2011-02-28 05:39:39Z oh2gve $
 * 
 * Copyright 2009, 2010, 2011 Tapio Aaltonen
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// own stuff.
#include "../serialport.h"

// std stuff.
#include <string>
using std::string;
#include <fstream>
using std::ofstream;
#include <ios>
#include <sstream>
using std::ostringstream;
#include <cstdio>
#include <cstring>
#include <cerrno>

// ye olde stuff.
#include <fcntl.h>   // File control definitions.
#include <termios.h> // POSIX terminal control definitions.
#include <sys/stat.h> // Used to see wether lock file exists.



Serialport::Serialport(string const& deviceFile, string const& lockFile, unsigned int const& timeout,
			unsigned int const& baudrate, unsigned int const& charsize, char const& parity, unsigned int const& stopbits, bool const& rtscts):
timeout_(timeout), error_(), portHandle_(-1), lockFileName_(lockFile)
{
	// Check if lock file exists.
	struct stat lfinfo;
	if ( stat(lockFile.c_str(), &lfinfo) == 0 )
	{
		// We got some info from it, so it must be there.
		error_ = string("it seems that ") + deviceFile + string(" is already in use, at least ") + lockFile + string(" exists.");
		
		// Clear lock file name so that the file won't be deleted at destructor.
		lockFileName_ = "";
		return;
	}
	else
	{
		// Good, it seems that it doesn't. Let's create it.
		ofstream lckf ( lockFile.c_str() );
		if ( !lckf )
		{
			// Bad, we can't create it.
			error_ = string("failed to create ") + lockFile + string(".");
			return;
		}
		else
		{
			lckf << getpid() << "\n";
		}
		lckf.close();
		
		// Save the name so that we can close it upon destruction.
		lockFileName_ = lockFile;
	}
	
	// Open the device.
	portHandle_ = open(deviceFile.c_str(), O_RDWR | O_NOCTTY | O_NDELAY);
	
	// Check for error.
	if ( portHandle_ == -1 )
	{
		error_ = string("failed to open ") + deviceFile + string(" (") + string(strerror(errno)) + string(")");
		return;
	}
	else
	{
		// Configure the device blocking.
		fcntl(portHandle_, F_SETFL, 0);
		
		// Get current config.
		struct termios cfg;
		tcgetattr(portHandle_, &cfg);
		
		// Enable local mode and receiver.
		cfg.c_cflag |= (CLOCAL | CREAD);
		
		// Select new baudrate.
		switch ( baudrate )
		{
			case 300:
				cfsetispeed(&cfg, B300);
				cfsetospeed(&cfg, B300);
				break;
			case 600:
				cfsetispeed(&cfg, B600);
				cfsetospeed(&cfg, B600);
				break;
			case 1200:
				cfsetispeed(&cfg, B1200);
				cfsetospeed(&cfg, B1200);
				break;         
			case 2400:
				cfsetispeed(&cfg, B2400);
				cfsetospeed(&cfg, B2400);
				break;
			case 4800:
				cfsetispeed(&cfg, B4800);
				cfsetospeed(&cfg, B4800);
				break;
			case 9600:
				cfsetispeed(&cfg, B9600);
				cfsetospeed(&cfg, B9600);
				break;
			case 19200:
				cfsetispeed(&cfg, B19200);
				cfsetospeed(&cfg, B19200);
				break;
			case 38400:
				cfsetispeed(&cfg, B38400);
				cfsetospeed(&cfg, B38400);
				break;
			case 57600:
				cfsetispeed(&cfg, B57600);
				cfsetospeed(&cfg, B57600);
				break;
			default:
				ostringstream errmsg;
				errmsg << "invalid baudrate " << baudrate;
				error_ = errmsg.str();
				return;
		}
		
		// Select parity.
		switch ( parity )
		{
			case 'N':
				cfg.c_cflag &= ~PARENB;
				cfg.c_iflag &= ~(INPCK | ISTRIP);
				break;
			case 'E':
				cfg.c_cflag |= PARENB;
				cfg.c_cflag &= ~PARODD;
				cfg.c_iflag |= (INPCK | ISTRIP);
				break;
			case 'O':
				cfg.c_cflag |= PARENB;
				cfg.c_cflag |= PARODD;
				cfg.c_iflag |= (INPCK | ISTRIP);
				break;
			default:
				error_ = string("invalid parity: ") + parity;
				return;
		}
		
		// Select amount of databits.
		cfg.c_cflag &= ~CSIZE;
		switch ( charsize )
		{
			case 7:
				cfg.c_cflag |= CS7;
				break;
			case 8:
				cfg.c_cflag |= CS8;
				break;
			default:
				ostringstream errmsg;
				errmsg << "invalid character size " << charsize;
				error_ = errmsg.str();
				return;
		}
		
		// Select stopbit count.
		switch ( stopbits )
		{
			case 1:
				cfg.c_cflag &= ~CSTOPB;
				break;
			case 2:
				cfg.c_cflag |= CSTOPB;
				break;
			default:
				ostringstream errmsg;
				errmsg << "invalid stopbit count: " << stopbits;
				error_ = errmsg.str();
				return;
		}
		
		// Select hw flow control status.
		if ( rtscts )
		{
			cfg.c_cflag |= CRTSCTS;
		}
		else
		{
			cfg.c_cflag &= ~CRTSCTS;
		}
		
		// Select raw input.
		cfg.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);
		
		// Disable sw flow control.
		cfg.c_iflag &= ~(BRKINT | INLCR | ICRNL | IXON | IXOFF | IXANY);
		
		// Select raw output.
		cfg.c_oflag &= ~OPOST;
		
		// Set read timeout, if requested.
		if ( timeout_ > 0 )
		{
			cfg.c_cc[VMIN] = 0;
			cfg.c_cc[VTIME] = timeout;
		}
		
		// Set new config.
		tcsetattr(portHandle_, TCSANOW, &cfg);
	}
	
}



Serialport::~Serialport()
{
	if ( !lockFileName_.empty() )
	{
		remove(lockFileName_.c_str());
	}
	if ( portHandle_ != -1 )
	{
		close(portHandle_);
	}
}



bool Serialport::transmit(string const& data)
{
	// Check that we're ok.
	if ( !error_.empty() )
	{
		return false;
	}
	
	// Write given data.
	int n = write(portHandle_, data.c_str(), data.size());
	
	// Check for errors.
	if ( n < 0 )
	{
		error_ = string("failed to write (") + string(strerror(errno)) + string(")");
		return false;
	}
	else if ( n != static_cast<int>(data.size()) )
	{
		ostringstream errmsg;
		errmsg << "tried to write " << data.size() << " bytes, but wrote " << n;
		error_ = errmsg.str();
		return false;
	}
	
	return true;
}



bool Serialport::receive(string& data, unsigned int const max)
{
	// Check that we're ok.
	if ( !error_.empty() )
	{
		return false;
	}
	
	// Read from device.
	char* buffer = new char[max];
	int n = read(portHandle_, buffer, max);
	
	// Check for errors.
	if ( n < 0 )
	{
		if ( timeout_ == 0 || errno != EAGAIN )
		{
			error_ = string("failed to read (") + string(strerror(errno)) + string(")");
			delete[] buffer;
			return false;
		}
	}
	else
	{
		// Save results.
		data = string(buffer, n);
	}
	
	delete[] buffer;
	return true;
}



bool Serialport::operator!() const
{
	return !error_.empty();
}



string Serialport::getErrorMessage() const
{
	return error_;
}
