/* $Id: packetlog.cpp 439 2011-02-16 20:55:02Z oh2gve $
 * 
 * Copyright 2007, 2008, 2009, 2010, 2011 Tapio Aaltonen
 *
 * 
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/

// own stuff.
#include "packetlog.h"
#include "cfg.h"
using Cfg::LogCfg;
#include "logserver.h"
#include "logfile.h"
#include "packet.h"

// std stuff.
#include <queue>
using std::queue;

// wx stuff.
#include <wx/thread.h>
#include <wx/string.h>
#include <wx/log.h>


Packetlog::Packetlog(LogCfg const& cfg, wxString const& ownCall):
wxThread(wxTHREAD_JOINABLE),
mutex_(), initialized_(false), keepgoing_(true), shouldReopen_(false), cfg_(cfg), ownCall_(ownCall),
buffer_(), file_(NULL), server_(NULL)
{
}


Packetlog::~Packetlog()
{
	while ( !buffer_.empty() )
	{
		delete buffer_.front();
		buffer_.pop();
	}
}


wxThread::ExitCode Packetlog::Entry()
{
	// Protect our startup from intrusions of main thread.
	mutex_.Lock();
	
	// This signals waitForInit() to go waiting on mutex_.
	initialized_ = true;
	
	// If something is about to happen, let user know.
	if ( cfg_.serverNeeded || cfg_.fileNeeded )
	{
		wxLogMessage(wxT("Starting packet logging."));
	}
	
	// Try file first.
	if ( cfg_.fileNeeded )
	{
		file_ = new Logfile(cfg_.packetFileName.ToAscii());
	
		// Check for fatal errors.
		if ( !(*file_) )
		{
			wxLogError(wxT("File ") + cfg_.packetFileName + wxT(" could not be opened."));
			keepgoing_ = false;
		}
		else
		{
			wxLogMessage(wxT("Packet log file opened."));
		}
	}
	
	// If file was successfull, try server.
	if ( keepgoing_ && cfg_.serverNeeded )
	{
		server_ = new LogServer(cfg_.packetSrv, ownCall_);
		
		// Check for fatal errors.
		if ( !(*server_) )
		{
			wxLogError(wxT("Log server could not be started."));
			keepgoing_ = false;
		}
		else
		{
			wxLogMessage(wxT("Log server started."));
		}
	}
	
	// Initialization complete, let's allow main to poll our operator!().
	mutex_.Unlock();
	
	// Main loop.
	while ( keepGoing() )
	{
		// Packets are written once in second.
		Sleep(1000);
		
		// Check statuses.
		if ( file_ != NULL && !(*file_) )
		{
			delete file_;
			file_ = NULL;
			wxLogMessage(wxT("Log server stopped."));
		}
		if ( server_ != NULL && !(*server_) )
		{
			delete server_;
			server_ = NULL;
			wxLogMessage(wxT("Packet log file closed."));
		}
		
		// Get exclusive access to flags.
		mutex_.Lock();
		
		// Check for need to reopen.
		if ( shouldReopen_ )
		{
			if ( cycleFile() )
			{
				wxLogMessage(wxT("Packet log file reopened."));
				shouldReopen_ = false;
			}
			else
			{
				wxLogError(wxT("Failed to reopen packet log file."));
				keepgoing_ = false;
			}
		}
		
		// Allow main to access again.
		mutex_.Unlock();
		
		// If server is enabled, check for new connections.
		if ( server_ != NULL )
		{
			server_->accept();
		}
		
		// Check for incoming packets. Wait for one second if there's none.
		mutex_.Lock();
		while ( !buffer_.empty() )
		{
			actualWrite(buffer_.front());
			buffer_.pop();
		}
		mutex_.Unlock();
	}
	
	// Cleanup.
	if ( server_ != NULL )
	{
		delete server_;
		server_ = NULL;
	}
	if ( file_ != NULL )
	{
		delete file_;
		file_ = NULL;
	}
	
	wxLogDebug(wxT("Packet logging thread is terminating."));

	// Thread terminates here.
	return NULL;
}


void Packetlog::waitForInit()
{
	while ( !initialized_ )
	{
		// Main thread does not need CPU now.
		Yield();
	}
	
	// Packet log thread is has started and is initializing. Let's wait for it to complete.
	mutex_.Lock();
	
	// Okay, initialization complete.
	mutex_.Unlock();
}


bool Packetlog::operator!() const
{
	return !keepgoing_;
}



void Packetlog::terminate()
{
	mutex_.Lock();
	keepgoing_ = false;
	mutex_.Unlock();
}



void Packetlog::write(Packet const* packet)
{
	if ( packet == NULL )
	{
		return;
	}
	
	// Store packet.
	mutex_.Lock();
	buffer_.push(packet);
	mutex_.Unlock();
}


void Packetlog::reopen()
{
	// Reserve access for main.
	mutex_.Lock();
	
	// Toggle flag.
	shouldReopen_ = true;
	
	// Release access back.
	mutex_.Unlock();
}





void Packetlog::actualWrite(Packet const* packet)
{
	// Write the line.
	if ( file_ != NULL )
	{
		file_->write(packet->toString(false));
	}
	if ( server_ != NULL )
	{
		server_->write(packet->toString(true));
	}
	
	// Packet has been logged, so let's erase it.
	delete packet;
}



bool Packetlog::keepGoing()
{
	bool retval;
	mutex_.Lock();
	retval = keepgoing_;
	mutex_.Unlock();
	return retval;
}

	

bool Packetlog::cycleFile()
{
	if ( cfg_.fileNeeded )
	{
		delete file_;
		file_ = new Logfile(cfg_.packetFileName.ToAscii());
		if ( !(*file_) )
		{
			delete file_;
			file_ = NULL;
			return false;
		}
	}
	
	return true;
}
