/* $Id: serialport.cpp 454 2011-02-27 10:30:45Z oh2gve $
 * 
 * Copyright 2010, 2011 Tapio Aaltonen
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// own stuff.
#include "../serialport.h"

// mingw stuff.
#include <windows.h>

// std stuff.
#include <string>
using std::string;
#include <sstream>
using std::ostringstream;



Serialport::Serialport(string const& deviceFile, unsigned int const& timeout,
			unsigned int const& baudrate, unsigned int const& charsize, char const& parity, unsigned int const& stopbits, bool const& rtscts):
timeout_(timeout), error_(), portHandle_(INVALID_HANDLE_VALUE)
{
	// Validate baudrate.
	switch ( baudrate )
	{
		case 300:
		case 600:
		case 1200:
		case 2400:
		case 4800:
		case 9600:
		case 19200:
		case 38400:
		case 57600:
			break;
		default:
			ostringstream errmsg;
			errmsg << "invalid baudrate " << baudrate;
			error_ = errmsg.str();
			return;
	}

	// Validate charsize.
	switch ( charsize )
	{
		case 7:
		case 8:
			break;
		default:
			ostringstream errmsg;
			errmsg << "invalid character size " << charsize;
			error_ = errmsg.str();
			return;
	}
	
	// Validate timeout.
	if ( timeout >= MAXDWORD )
	{
		ostringstream errmsg;
		errmsg << "invalid timeout " << timeout;
		error_ = errmsg.str();
	}

	// Prepare to open device.
	LPCSTR portname = deviceFile.c_str();
	DWORD accessdirection = GENERIC_READ | GENERIC_WRITE;
	
	// Open device.
	portHandle_ = CreateFile(portname, accessdirection, 0, 0, OPEN_EXISTING, 0, 0);
	
	// Check how we did.
	if ( portHandle_ == INVALID_HANDLE_VALUE )
	{
		ostringstream errmsg;
		errmsg << string("failed to open ") << deviceFile << string(" (") << GetLastError() << string(")");
		error_ = errmsg.str();
		return;
	}

	// Get current port config.
	DCB dcbSerialParams = {0};
	dcbSerialParams.DCBlength = sizeof(dcbSerialParams);
	
	if ( !GetCommState(portHandle_, &dcbSerialParams) )
	{
		error_ = string("failed to get initial state of") + deviceFile;
		return;
	}
	
	// Set our parameters as config.
	dcbSerialParams.BaudRate = baudrate;
	dcbSerialParams.ByteSize = charsize;
	switch ( stopbits )
	{
		case 1:
			dcbSerialParams.StopBits = ONESTOPBIT;
			break;
		case 2:
			dcbSerialParams.StopBits = TWOSTOPBITS;
			break;
		default:
			ostringstream errmsg;
			errmsg << "invalid stopbit count: " << stopbits;
			error_ = errmsg.str();
			return;
	}
	switch ( parity )
	{
		case 'N':
			dcbSerialParams.Parity = NOPARITY;
			break;
		case 'E':
			dcbSerialParams.Parity = EVENPARITY;
			break;
		case 'O':
			dcbSerialParams.Parity = ODDPARITY;
			break;
		default:
			error_ = string("invalid parity: ") + parity;
			return;
	}
	
	// Set our parameters into port.
	if( !SetCommState(portHandle_, &dcbSerialParams) )
	{
		ostringstream errmsg;
		errmsg << string("failed to configure ") << deviceFile << string(" (") << GetLastError() << string(")");
		error_ = errmsg.str();
	}
	
	// Prepare to set timeouts.
	COMMTIMEOUTS timeouts = {0};
	timeouts.ReadIntervalTimeout = MAXDWORD;
	timeouts.ReadTotalTimeoutConstant = timeout;
	timeouts.ReadTotalTimeoutMultiplier = MAXDWORD;
	timeouts.WriteTotalTimeoutConstant = 0;
	timeouts.WriteTotalTimeoutMultiplier = 0;

	// Set timeouts.
	if( !SetCommTimeouts(portHandle_, &timeouts) )
	{
		ostringstream errmsg;
		errmsg << string("failed to set I/O timeouts of ") << deviceFile << string(" (") << GetLastError() << string(")");
		error_ = errmsg.str();
	}
}



Serialport::~Serialport()
{
	if ( portHandle_ != INVALID_HANDLE_VALUE )
	{
		CloseHandle(portHandle_);
	}
}



bool Serialport::transmit(string const& data)
{
	// Check that we're ok.
	if ( !error_.empty() )
	{
		return false;
	}
	
	DWORD dwBytesWritten = 0;
	if ( !WriteFile(portHandle_, data.c_str(), data.size(), &dwBytesWritten, NULL) )
	{
		ostringstream errmsg;
		errmsg << string("write failed ") << string(" (") << GetLastError() << string(")");
		error_ = errmsg.str();
		
	}

	if ( dwBytesWritten != data.size() )
	{
		ostringstream errmsg;
		errmsg << "tried to write " << data.size() << " bytes, but wrote " << dwBytesWritten;
		error_ = errmsg.str();
		return false;
	}
	
	return true;
}



bool Serialport::receive(string& data, unsigned int const max)
{
	// Check that we're ok.
	if ( !error_.empty() )
	{
		return false;
	}

	// Prepare buffers.
	data.clear();
	char* buffer = new char[max];
	
	// Read data.
	DWORD dwBytesRead = 0;
	if ( !ReadFile(portHandle_, buffer, max, &dwBytesRead, NULL) )
	{
		ostringstream errmsg;
		errmsg << string("read failed ") << string(" (") << GetLastError() << string(")");
		error_ = errmsg.str();
		delete[] buffer;
		return false;
	}
	else
	{
		// Store data.
		data = string(buffer, dwBytesRead);
	}

	delete[] buffer;
	return true;
}



bool Serialport::operator!() const
{
	return !error_.empty();
}



string Serialport::getErrorMessage() const
{
	return error_;
}
